// modsplit is a tool to create redundant backups of your files
// written 2002-11 by Thorsten Reinecke
// Version 0.9d, last change: 2002-11 by Thorsten Reinecke
// Copyright (C) 2002 by Thorsten Reinecke (reinecke@thorstenreinecke.de)

/*
    This program is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
    Version 2 (June 1991). See the "COPYING" file distributed with this
    software for more info.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/


// short summary:
// modsplit uses simultaneous congruences to create up to 8 files out of one
// data file. Each of these files has about 25.3% of the original filesize.
// Only four out of these files are necessary to reconstruct your original
// data file. This gives you up to 70 different possibilities to restore
// your data.
// Let's take a short example:
// 2766 Megabytes of Data can be splitted to six 700MB CDs, and even if
// two out of these CDs get lost, you will be able to restore your
// 2766 Megabytes. I would consider this as a really safe backup!
//
// *** WARNING: ***
// As this program may have errors, you have to check that you are really 
// able to restore your data. For this reason you should try to restore
// your data after you have splitted it!
//
// Disclaimer:
// Under no circumstances I will take any responsibility for loss of data
// or any other harm which is caused by using this program!
// If algorithms used in this program are covered by legal patents, and these
// patents are valid in your country, then you are not allowed to use this
// program without permission of the patent holder. (One never knows, whether
// the Chinese Remainder Theorem is patented anywhere...)
// If you are not willing to accept these terms of usage, then you are not
// allowed to use this program!


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <getopt.h>


const char* const option_string = "+hvs:f:r:";
static const struct option long_options[] =
{
  { "help", 0, 0, 'h' },
  { "verbose", 0, 0, 'v'},
  { "split", 1, 0, 's' },
  { "first", 1, 0, 'f' },
  { "restore", 1, 0, 'r' },
  { 0, 0, 0, 0 }
}; // refer "man 3 getopt" for this struct

void usage(const char* const prg)
{
  printf("modsplit is a tool to create redundant backups of your files\n");
  printf("(C) 2002 by Thorsten Reinecke (reinecke@thorstenreinecke.de)\n");
  printf("Version 0.9d, distributed under GNU Public License Version 2\n");
  printf("short summary:\n");
  printf("modsplit uses simultaneous congruences to create up to 8 files out of one\n");
  printf("data file. Each of these files has about 25.3%% of the original filesize.\n");
  printf("Only four out of these files are necessary to reconstruct your original\n");
  printf("data file. This gives you up to 70 different possibilities to restore\n");
  printf("your data.\n");
  printf("usage: \n");
  printf("%s -s <file> [-f <1-8>] <1-8 splitfilenames>\n",prg);
  printf("%s -r <file> <4 splitfilenames>\n",prg);
  printf("  -s, --split '-'|<filename>      create a backup from stdin or <filename>\n");
  printf("    1-8 filenames for the destination files have to be specified.\n");
  printf("  -f, --first <1..8>              specify slot of first modulus\n");
  printf("    this is useful, if you want to split sequentially\n");
  printf("\n  -r, --restore '-'|<filename>    restore a backup to stdout or <filename>\n");
  printf("    exactly 4 filenames for splitted source files need to be given.\n\n");
  printf("  -v, --verbose            be verbose\n");
  printf("  -h, --help               this text\n");
}


enum Toperation_mode { op_undefined, op_split, op_restore };
int op_mode = op_undefined;
int verbose = 0;

int splits = 0;
int first_slot_to_use = 0; // this is the default

#define MAXSPLITS 8
#define RELBITS 31


const unsigned int M_moduli[MAXSPLITS] =
 { 239, 241, 245, 247, 248, 249, 251, 253 };
 
const unsigned int* moduli = M_moduli; // may be set to M_moduli[first_slot_to_use]

#define RELBITS_MAX 32
#define MPRODUCT (239u*241u*245u*247u)
#define BITPOT (1u<<RELBITS)
const unsigned int max_unique = MPRODUCT-1;
const unsigned int min_unique = BITPOT-1;
const unsigned int low_unique = MPRODUCT-BITPOT-1;

// 0..min_unique is the range that must match, no matter what
// numbers we have chosen.
// Due to the Chinese Ramainder Theorem, the range can be expanded
// to 0..max_unique.
// In the subrange min_unique+1..max_unique we save one bit! This is the
// reason, why we do not maximize the number of moduli. (We would be
// able to go just below 2^(31/4), which is about 215, and which would give
// us 14 disjoint numbers instead of 8. But who needs so much redundancy?)
// If we use this range, we are able to save one bit in the subrange
// 0..low_unique, too.


unsigned int restore_moduli[3] = { 0,0,0 }; // to be filled later

const char magic_cookie[] = "modsplit 8/31 file";


#define my_FILEMODE_READ  00000
#define my_FILEMODE_WRITE 00666

const char* origfile = NULL;
int fd_origfile;
int fd_splittedfile[MAXSPLITS];
off_t origfilesize;

#define bufsize 4096
unsigned char splitbuf[MAXSPLITS][bufsize];
int splbufpos = 0; 


void wegschreiben(unsigned int bits, int flush)
{
 int i;
 for (i=0; i<splits; ++i) splitbuf[i][splbufpos]=bits%moduli[i];
 if (++splbufpos==bufsize || flush)
  {

#if 0
   // This is yet another possibility to save bits and reduce the
   // overhead of our transformation...
   // remember: we use subranges 0..m<=256 instead of 0..255!
   // This method is not used, because it would result in different
   // filesizes and more complicated buffer-handling.
   // HINT: Let gzip or other compression tools do this job afterwards,
   // they are simply more efficient!
   {
     static unsigned int hiddenbits = 0; 
     int i,j;
     for (i=0; i<splits; ++i)
      for (j=0; j<splbufpos; ++j)
       hiddenbits+=(splitbuf[i][j]<256-moduli[i]);
     sprintf(stderr,"Hidden bytes: %u ...\n",hiddenbits/8);
   }
#endif  
  
   for (i=0; i<splits; ++i) write(fd_splittedfile[i],&splitbuf[i],splbufpos);
   splbufpos=0;
  }
}

void do_split(void)
{
 int i;
 int rsize = 0;
 unsigned char rbuf[bufsize];
 unsigned int bits=0;
 unsigned int bitpos=0;
 
 if (verbose)
  {
    fprintf(stderr,"split using %i",moduli[0]);
    for (i=1; i<splits; ++i) fprintf(stderr,", %i",moduli[i]);
    fprintf(stderr,"\n");
  }

 // do the splitting
 
 while ( (rsize=read(fd_origfile,rbuf,bufsize)) )
  {
    if (rsize<0)
     {
       if (verbose) perror(NULL);
       exit(2);
     }
    //printf("%i bytes read from %s.\n",rsize,origfile);
    for (i=0; i<rsize; ++i)
     {
       unsigned int h = rbuf[i];
       bits|=h<<bitpos; bitpos+=8;
       if (bitpos>=RELBITS_MAX)
        {
          unsigned int bits_to_store = RELBITS;
          if ( bits<=low_unique || (bits>min_unique && bits<=max_unique) )
           {
             bits_to_store=RELBITS_MAX;
             wegschreiben(bits,0);
             if (RELBITS_MAX>=32) bits=0; // because "bits>>=32" seems to result in identity instead of 0!
           }
          else wegschreiben(bits&min_unique,0);
          
          if (bitpos>32)
           {
             // printf("overflow A: bits: %u pos: %u stored_bits: %u\n",bits,bitpos,bits_to_store);
             h>>=8-(bitpos&7u);
             bitpos-=bits_to_store; bits>>=bits_to_store;
             bits|=h<<(32-bits_to_store);
             // printf("overflow B: bits: %u pos: %u stored_bits: %u\n",bits,bitpos,bits_to_store);
           }
          else
           {
             bitpos-=bits_to_store; bits>>=bits_to_store;
           }
          // printf("bits: %u pos: %u\n",bits,bitpos);
        }
     }
  }
 if (bitpos) wegschreiben(bits,1);


 // close the files
 for (i=0; i<splits; ++i) close(fd_splittedfile[i]);
 close(fd_origfile);
}


unsigned int invmod(unsigned int x, const unsigned int m)
{
  // returns 1/x mod m (or error if no inverse exists) 
  unsigned int a,b,q,r;
  signed int h,inv,sw;
  inv=1; sw=0;
  a=x%=m; b=m;
  while (b)
   {
     q=a/b; r=a%b;
     a=b; b=r;
     h=inv-sw*q; inv=sw; sw=h;
   }
  if (a!=1) exit(1); // inverse does not exist!
  a = (inv>=0) ? inv : m+inv;
  return a;
}


unsigned int modquot(unsigned int r1, unsigned int r2, unsigned int inv, unsigned int m)
{
 if (r1>=r2)
  return ((r1-r2)*inv)%m;
 else
  return m-(((r2-r1)*inv)%m);
}                           

void do_restore(void)
{
 int i,j;
 int wsize = 0;
 unsigned char wbuf[bufsize];
 unsigned int bits=0;
 unsigned int bitpos=0;
 
 const unsigned int
   inv_m3_m0 = invmod(restore_moduli[3],restore_moduli[0]),
   inv_m3_m1 = invmod(restore_moduli[3],restore_moduli[1]),
   inv_m3_m2 = invmod(restore_moduli[3],restore_moduli[2]),
   inv_m2_m0 = invmod(restore_moduli[2],restore_moduli[0]),
   inv_m2_m1 = invmod(restore_moduli[2],restore_moduli[1]),
   inv_m1_m0 = invmod(restore_moduli[1],restore_moduli[0]);
   
 const unsigned int
   m012 = restore_moduli[0]*restore_moduli[1]*restore_moduli[2];
   
 const unsigned int
   m0123 = m012*restore_moduli[3];
   
 int b_read[4];

 if (!restore_moduli[0] || !restore_moduli[1] || !restore_moduli[2] || !restore_moduli[3])
  {
    fprintf(stderr,"not enough splitted files to restore data!\n");
    unlink(origfile);
    exit(1);
  }
  
 for (i=0; i<4; ++i) for (j=i+1; j<4; ++j) // check for distinct moduli
  if (restore_moduli[i]==restore_moduli[j])
   {
     fprintf(stderr,"splitted files must be different to restore data!\n");
     unlink(origfile);
     exit(1);
   }

 if (verbose)
  {
    fprintf(stderr,"restore using %i, %i, %i, %i\n",restore_moduli[0],restore_moduli[1],restore_moduli[2],restore_moduli[3]);
  }    

 // do restore
 
 while (1)
  {
    int j,k;
    for (j=0; j<4; ++j) b_read[j]=read(fd_splittedfile[j],&splitbuf[j],bufsize);
    
    for (j=0; j<4; ++j) for (k=j+1; k<4; ++k) // read check
     if (b_read[j]!=b_read[k])
     {
       fprintf(stderr,"read-error, giving up!\n");
       exit(2);
     }
    if (b_read[0]==0) break; // nothing left...
    for (i=0; i<b_read[0]; ++i)      
     {
       const unsigned char
        r0=splitbuf[0][i],
        r1=splitbuf[1][i],
        r2=splitbuf[2][i],
        r3=splitbuf[3][i];
       
       const unsigned char
        s0=modquot(r0,r3,inv_m3_m0,restore_moduli[0]),
        s1=modquot(r1,r3,inv_m3_m1,restore_moduli[1]),
        s2=modquot(r2,r3,inv_m3_m2,restore_moduli[2]);
        
       const unsigned int
        t0=modquot(s0,s2,inv_m2_m0,restore_moduli[0]),
        t1=modquot(s1,s2,inv_m2_m1,restore_moduli[1]);
        
       const unsigned int
        u0=modquot(t0,t1,inv_m1_m0,restore_moduli[0]);
        
       const unsigned int z=u0*restore_moduli[1]+t1;
       const unsigned int y=(z*restore_moduli[2]+s2)%(m012);
       const unsigned int x=(y*restore_moduli[3]+r3)%(m0123);
       
       unsigned int bits_to_store = (x<=low_unique) ?
        RELBITS_MAX : (x<=min_unique) ? RELBITS : RELBITS_MAX;
       // 0            .. low_unique : RELBITS_MAX
       // low_unique+1 .. min_unique : RELBITS
       // min_unique+1 .. max_unique : RELBITS_MAX
       // >max_unique                : error!!
       
       // to avoid risk of bit overflow, we split our shift-operation
       // nasty cut & paste, but hey: who cares?
       unsigned int lox = x&0xffffU;
       unsigned int hix = x>>16;

       if (x>max_unique)
        {
          printf ("%u %u %u\n",x,y,z);
          printf("sanity check failed! x=%u\n",x); exit(2);
        }

       bits|=lox<<bitpos; bitpos+=16;
       while (bitpos>7)
        {
          wbuf[wsize]=bits; --origfilesize;
          if (++wsize==bufsize || !origfilesize)
           {
             write(fd_origfile,wbuf,wsize); wsize=0;
             // printf("remain: %u \n",origfilesize);
           }
          bits>>=8; bitpos-=8;
          if (!origfilesize) goto fertig;
        }

       bits|=hix<<bitpos; bitpos+=bits_to_store-16;
       while (bitpos>7)
        {
          wbuf[wsize]=bits; --origfilesize;
          if (++wsize==bufsize || !origfilesize)
           {
             write(fd_origfile,wbuf,wsize); wsize=0;
             // printf("remain: %u \n",origfilesize);
           }
          bits>>=8; bitpos-=8;
          if (!origfilesize) goto fertig;
        }

     }
  }
  
 if (origfilesize)
  {
    wbuf[wsize]=bits;
    ++wsize;
  }
 if (wsize) { write(fd_origfile,wbuf,wsize); wsize=0; }

fertig:
 // close the files
 for (i=0; i<splits; ++i) close(fd_splittedfile[i]);
 close(fd_origfile);
}


int main(int argc, char* argv[])
{
  int option_index;
  int o;
  while ( (o=getopt_long(argc,argv,option_string,long_options,&option_index)) != -1)
   {
     switch (o)
      {
        case 'h': usage(argv[0]); exit(0); break;
        case 'v': verbose=1; break;
        case 's': op_mode=op_split; origfile=strdup(optarg);
                  if (strcmp(origfile,"-")==0) fd_origfile=dup(0); // for pipe
                  else
                   if ((fd_origfile=open(optarg, O_RDONLY, my_FILEMODE_READ))==-1)
                    {
                      fprintf(stderr,"error while opening original file %s!\n",optarg);
                      exit(1);
                    }
                  {
                    struct stat mybuf;
                    fstat(fd_origfile,&mybuf);
                    origfilesize=mybuf.st_size;
                  }  
                  break;
        case 'f': first_slot_to_use=atoi(optarg)-1;
                  if (first_slot_to_use<0 || first_slot_to_use>=MAXSPLITS)
                   {
                     fprintf(stderr,"first must be between 1 and %i\n",MAXSPLITS);
                     exit(1);
                   }
                  else moduli=&M_moduli[first_slot_to_use];
                  break;
        case 'r': op_mode=op_restore; origfile=strdup(optarg);
                  if (strcmp(origfile,"-")==0) fd_origfile=dup(1); // for pipe
                  else
                   if ((fd_origfile=open(optarg, O_WRONLY|O_CREAT|O_EXCL, my_FILEMODE_WRITE))==-1)
                    {
                      fprintf(stderr,"error while creating original file %s!\n",optarg);
                      exit(1);
                    }
                  break;
       default: usage(argv[0]); exit(1);
      }
   }
  // now optind points to the first argv-element that is not an option (refer man getopt)

  if (op_mode==op_undefined)
   {
     fprintf(stderr," undefined operation mode. please use --help\n");
     exit(1);
   }

  if (op_mode==op_restore)
   {
    if (optind+4!=argc)
     {
       fprintf(stderr,"We need exactly four filenames to restore a file!\n");
       unlink(origfile);
       exit(1);
     }
    for (splits=0; splits<4; ++splits)
     {
       const int p=splits;
       char* optarg=argv[optind+splits];
       char my_cookie[sizeof(magic_cookie)];
       
       if ((fd_splittedfile[p]=open(optarg, O_RDONLY, my_FILEMODE_READ))==-1)
        {
         fprintf(stderr,"error while opening %s!\n",optarg);
         unlink(origfile);
         exit(1);
        }

       bzero(my_cookie,sizeof(magic_cookie));
       read(fd_splittedfile[p],my_cookie,sizeof(magic_cookie));
       if (strcmp(my_cookie,magic_cookie))
        {
          fprintf(stderr,"error: %s is not a splitted backup file!\n",optarg);
          unlink(origfile);
          exit(1);
        }
       read(fd_splittedfile[p],&restore_moduli[p],sizeof(restore_moduli[p]));
       read(fd_splittedfile[p],&origfilesize,sizeof(origfilesize));
     }
   }

  if (op_mode==op_split)
   {
    if (optind+1>argc)
     {
       fprintf(stderr,"We need at least one filename to split a file!\n");
       exit(1);
     }
    if (optind+MAXSPLITS-first_slot_to_use<argc)
     {
       fprintf(stderr,"In this configuration we need at most %i filenames to split a file!\n",MAXSPLITS-first_slot_to_use);
       exit(1);
     }
    if (verbose && optind+4>argc)
     {
       fprintf(stderr,"WARNING: I'll do it, but you need four splits for restoring your data!\n");
     }
     
    for (splits=0; splits<MAXSPLITS; ++splits)
     {
       const int i=splits;
       const char* filename=argv[optind+splits];
       if (optind+splits>=argc) break;

       if ((fd_splittedfile[i]=open(filename, O_WRONLY|O_CREAT|O_EXCL, my_FILEMODE_WRITE))==-1)
        {
          int j;
          fprintf(stderr,"error while creating splitted file %s!\n",filename);
          for (j=0; j<splits; ++j)
           {
             close(fd_splittedfile[j]);
             unlink(argv[optind+j]);
           }
          exit(1);
        }
       write(fd_splittedfile[i],magic_cookie,sizeof(magic_cookie));
       write(fd_splittedfile[i],&moduli[i],sizeof(moduli[i]));
       write(fd_splittedfile[i],&origfilesize,sizeof(origfilesize));
     }
  }

 if (op_mode==op_split) do_split();
 if (op_mode==op_restore) do_restore();
 
 return 0;
}
